"""Configuration for covcheck validation."""

from pathlib import Path
from typing import Any, Optional, Union

from covcheck._cli.utilities import fail_with_error

try:
    import toml
    TOML_INSTALLED = True
except ImportError:
    TOML_INSTALLED = False


class Config:
    """Configuration for covcheck validation."""
    def __init__(
        self,
        coverage_filepath: Union[str, Path],
        line: Optional[float] = None,
        branch: Optional[float] = None,
        output: Optional[Union[str, Path]] = None,
        silent: bool = False,
    ):
        """Create a covcheck config.

        :param coverage_filepath: XML coverage file from Coverage.py.
        :param line: Threshold for line coverage.
        :param branch: Threshold for branch coverage.
        :param output: Path on disk where a JSON output file should be saved.
        :param silent: Whether the actaual coverage measurements should be printed to stdout.
        """
        self.coverage_filepath = coverage_filepath
        self.line = line
        self.branch = branch
        self.output = output
        self.silent = silent

    @classmethod
    def create(
        cls,
        coverage_filepath: Union[str, Path],
        config_filepath: Optional[Union[str, Path]] = None,
        group: Optional[str] = None,
        **kwargs: Optional[Any],
    ) -> 'Config':
        """Get a config based on both CLI arguments and config file settings.

        If the same argument is specified in both CLI arguments and config file,
        then the CLI argument will take precedence.

        :param coverage_filepath: Path to XML coverage file from Coverage.py.
        :param config_filepath: Path to pyproject.toml config file.
        :param group: Name of coverage group to check.
        :param kwargs: Kwargs containing additional config settings.
        """
        config = cls(coverage_filepath)

        if config_filepath is not None:
            if not TOML_INSTALLED:
                raise ImportError("--config was passed, but the toml package was not installed. "
                                  "Please 'pip install toml' before running with --config.")

            with open(config_filepath, 'r', encoding='utf-8') as f:
                toml_config = toml.load(f)

                if 'tool' in toml_config:
                    toml_tools = toml_config['tool']
                    if 'covcheck' in toml_tools:
                        toml_covcheck = toml_tools['covcheck']

                        # Check for missing group configuration
                        if group is not None:
                            # Group name selected but no matching group in config file
                            if 'group' in toml_covcheck and group not in toml_covcheck['group']:
                                fail_with_error(f"Group {group} not found in config")
                            # Group name selected but no groups at all in config file
                            if 'group' not in toml_covcheck:
                                fail_with_error(f"Group {group} not found in config")

                        if 'group' in toml_covcheck and group is not None:
                            group_dict = toml_covcheck['group'][group]
                            if 'coverage' in group_dict:
                                for key, value in group_dict['coverage'].items():
                                    setattr(config, key, value)
                        else:
                            for key, value in toml_covcheck.items():
                                setattr(config, key, value)

        for key, value in kwargs.items():
            if value is not None:
                setattr(config, key, value)

        return config



"""Coverage result."""

from pathlib import Path
from typing import Union

from covcheck._parsing.coverage_node import CoverageNode
from covcheck._parsing.coverage_summary import CoverageSummary
from covcheck._parsing.coverage_xml_parser import CoverageXMLParser


class CoverageResult:
    """Coverage result."""
    def __init__(self, tree: CoverageNode):
        self.tree = tree

    @property
    def summary(self) -> CoverageSummary:
        """Get the result CoverageSummary."""
        return self.tree.summary

    @classmethod
    def from_xml(cls, filepath: Union[str, Path]) -> 'CoverageResult':
        """Create a CoverageResult by parsing an XML coverage file.

        :param filepath: Path on disk to an XML coverage file.
        """
        tree = CoverageXMLParser.parse(filepath)
        return cls(tree)



"""Validation entrypoint."""

import json

from typing import Optional

from covcheck._parsing.coverage_result import CoverageResult
from covcheck._cli.utilities import fail_with_error
from covcheck._cli.config import Config


def validate_coverage(config: Config) -> None:
    """Validate code coverage given an XML coverage file.

    param: Config object.
    """
    _validate_thresholds(config.line, config.branch)

    result = CoverageResult.from_xml(config.coverage_filepath)

    if config.output is not None:
        with open(config.output, 'w', encoding='utf-8') as f:
            json.dump(result.tree.serialize(), f, indent=4)  # type: ignore

    required_args = [config.line, config.branch, config.output]
    if all(input_value is None for input_value in required_args):
        fail_with_error("Must specify --line, --branch, or --output_filepath.")

    checks_failed = False

    line_rate = result.summary.line_rate * 100
    if config.line is not None:
        if line_rate < config.line:
            fail_with_error(f"Line coverage ({line_rate:.2f}%) below threshold ({config.line}%)", sys_exit=False)
            checks_failed = True
        elif not config.silent:
            print(f"Line coverage passed: {line_rate:.2f}%")

    branch_rate = result.summary.branch_rate * 100
    if config.branch is not None:
        if branch_rate < config.branch:
            fail_with_error(f"Branch coverage ({branch_rate:.2f}%) below threshold ({config.branch}%)", sys_exit=False)
            checks_failed = True
        elif not config.silent:
            print(f"Branch coverage passed: {branch_rate:.2f}%")

    if checks_failed:
        fail_with_error("One or more quality checks failed.")


def _validate_thresholds(
    line_threshold: Optional[float],
    branch_threshold: Optional[float],
) -> None:
    """Validate thresholds are within acceptable bounds.

    :param line_threshold: Threshold for line coverage.
    :param branch_threshold: Threshold for branch coverage.
    """
    coverage_map = {
        'line': line_threshold,
        'branch': branch_threshold,
    }
    for coverage_type, threshold in coverage_map.items():
        if threshold is not None and (threshold < 0 or threshold > 100):
            fail_with_error(
                f"Invalid threshold for {coverage_type} coverage ({threshold}). Must be between 0 and 100.")



"""CLI entrypoint."""

import argparse

from covcheck._cli.config import Config
from covcheck._cli.validate import validate_coverage


def parse_args() -> argparse.Namespace:
    """Parse command-line arguments.

    :return: Argparse namespace containing the arguments.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('coverage_file', help="Path to XML coverage file.")
    parser.add_argument('--line', default=None, type=float, help="Line coverage percentage threshold.")
    parser.add_argument('--branch', default=None, type=float, help="Branch coverage percentage threshold.")

    parser.add_argument('--output', default=None, type=str, help="Path to a file where output JSON should be saved.")
    parser.add_argument('--silent', default=False, action='store_true', help="Do not print coverage results.")

    parser.add_argument('--config', default=None, type=str, help="Path to pyproject.toml config file.")
    parser.add_argument('--group', default=None, type=str, help="Name of coverage group to check.")

    return parser.parse_args()


def run() -> None:
    """Run the covcheck CLI."""
    args = parse_args()

    config = Config.create(
        args.coverage_file,
        config_filepath=args.config,
        line=args.line,
        branch=args.branch,
        output=args.output,
        silent=args.silent,
        group=args.group,
    )

    validate_coverage(config)



"""CLI utilities."""

import sys


class ColorCodes:
    """Color codes for color terminal printing"""
    FAIL = '\033[91m'
    BOLD = '\033[1m'
    END = '\033[0m'


def fail_with_error(
    message: str,
    sys_exit: bool = True,
    color: bool = True,
) -> None:
    """Print a colorized error message to the terminal.

    :param message: Message to print.
    :param sys_exit: Whether to exit the Python interpreter with sys.exit().
    :param color: Whether to write output in color.
    """
    full_message = message
    if color:
        full_message = f"{ColorCodes.FAIL}{ColorCodes.BOLD}{message}{ColorCodes.END}"

    print(full_message, file=sys.stderr)

    if sys_exit:
        sys.exit(1)



"""Covcheck main module."""

from covcheck._parsing.coverage_node import CoverageNode
from covcheck._parsing.coverage_node_type import CoverageNodeType
from covcheck._parsing.coverage_result import CoverageResult
from covcheck._parsing.coverage_summary import CoverageSummary

__all__ = [
    'CoverageNode',
    'CoverageNodeType',
    'CoverageResult',
    'CoverageSummary',
]



"""XML parser for coverage files."""

import re

from pathlib import Path
from typing import Union
from xml.etree import ElementTree
from xml.etree.ElementTree import Element

from covcheck._parsing.coverage_node import CoverageNode
from covcheck._parsing.coverage_node_type import CoverageNodeType
from covcheck._parsing.coverage_summary import CoverageSummary


class CoverageXMLParser:
    """XML parser for coverage files."""
    @classmethod
    def parse(cls, filepath: Union[str, Path]) -> CoverageNode:
        """Parse an XML coverage file into a covcheck tree.

        :param filepath: Path on disk to an XML coverage file.
        """
        xml_tree = ElementTree.parse(filepath)
        xml_root = xml_tree.getroot()

        root_node = CoverageNode('root', node_type=CoverageNodeType.DIR)

        xml_packages = cls._try_get_child(xml_root, 'packages')
        for xml_package in xml_packages:
            xml_classes = cls._try_get_child(xml_package, 'classes')
            for xml_class in xml_classes:
                code_filename = xml_class.attrib['name']
                full_filepath = xml_class.attrib['filename']
                code_dirpath = Path(full_filepath).parent if '/' in full_filepath else None

                summary = CoverageSummary(0, 0, 0, 0)
                xml_lines = cls._try_get_child(xml_class, 'lines')
                for xml_line in xml_lines:
                    summary.n_lines += 1
                    if xml_line.attrib['hits'] == '1':
                        summary.n_lines_covered += 1

                    if 'branch' in xml_line.attrib and xml_line.attrib['branch']:
                        branch_condition = xml_line.attrib['condition-coverage']
                        pattern = r"^\d+% \((\d+)\/(\d+)\)$"
                        match = re.match(pattern, branch_condition)

                        if match is None:
                            raise ValueError(f"Failed to parse condition-coverage XML: {branch_condition}")

                        summary.n_branches_covered += int(match.group(1))
                        summary.n_branches += int(match.group(2))

                node = CoverageNode(code_filename, node_type=CoverageNodeType.FILE, summary=summary)
                root_node.add_child(node, dirpath=code_dirpath)

        return root_node

    @classmethod
    def _try_get_child(cls, xml_element: Element, tag: str) -> Element:
        for xml_child in xml_element:
            if xml_child.tag == tag:
                return xml_child
        raise ValueError(f"Could not parse coverage XML, no attribute '{tag}'")



"""Coverage summary."""


class CoverageSummary:
    """Coverage summary."""
    def __init__(
        self,
        n_lines: int,
        n_lines_covered: int,
        n_branches: int,
        n_branches_covered: int,
    ):
        self.n_lines = n_lines
        self.n_lines_covered = n_lines_covered
        self.n_branches = n_branches
        self.n_branches_covered = n_branches_covered

    @property
    def line_rate(self) -> float:
        """Get the summary line rate."""
        if self.n_lines == 0:
            return 0
        return 1.0 * self.n_lines_covered / self.n_lines

    @property
    def branch_rate(self) -> float:
        """Get the summary branch rate."""
        if self.n_branches == 0:
            return 0
        return 1.0 * self.n_branches_covered / self.n_branches



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['covcheck', 'covcheck._cli', 'covcheck._parsing']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['covcheck = covcheck._cli.main:run']}

setup_kwargs = {
    'name': 'covcheck',
    'version': '0.4.2',
    'description': 'Code coverage validation',
    'long_description': '<div align="center">\n  <img src="https://storage.googleapis.com/hume-public-logos/covcheck/covcheck-dino-banner.png">\n  <h1>covcheck</h1>\n\n  <p>\n    <strong>Command-line tool for code coverage validation</strong>\n  </p>\n\n  <br>\n  <div>\n    <a href="https://badge.fury.io/py/covcheck"><img src="https://badge.fury.io/py/covcheck.svg" alt="PyPI"></a>\n    <a href="https://pepy.tech/project/covcheck"><img src="https://pepy.tech/badge/covcheck" alt="Downloads"></a>\n    <a href="https://github.com/HumeAI/covcheck/actions/workflows/ci.yml"><img src="https://github.com/HumeAI/covcheck/actions/workflows/ci.yaml/badge.svg" alt="CI"></a>\n  </div>\n  <br>\n</div>\n\n## About\n\n`covcheck` is intended to be used in conjunction with [coverage.py](https://coverage.readthedocs.io/), which already has support for `pytest`, `unittest`, and `nosetest`. All you have to do is point `covcheck` to the `coverage.xml` file produced when running your tests.\n\nFor more information on how to use `covcheck` please check out the [official docs page](https://humeai.github.io/covcheck).\n',
    'author': 'Hume AI Dev',
    'author_email': 'dev@hume.ai',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/HumeAI/covcheck',
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)



"""Coverage node."""

from pathlib import Path
from typing import Any, Dict, Generator, Optional, Union

from covcheck._parsing.coverage_summary import CoverageSummary
from covcheck._parsing.coverage_node_type import CoverageNodeType


class CoverageNode:
    """Coverage node."""
    def __init__(self, name: str, node_type: CoverageNodeType, summary: Optional[CoverageSummary] = None):
        """Construct CoverageNode.

        :param name: Name of the coverage node.
        :param node_type: Type of the coverage node.
        :param summary: Summary of coverage for the node.
        """
        self._name = name
        self._node_type = node_type
        self._summary = summary
        self._children = {}  # type: Dict[str, CoverageNode]

    @property
    def name(self) -> str:
        """Get the CoverageNode name.

        :return: CoverageNode name.
        """
        return self._name

    @property
    def node_type(self) -> CoverageNodeType:
        """Get the CoverageNode type.

        :return: CoverageNode type.
        """
        return self._node_type

    def add_child(self, node: 'CoverageNode', dirpath: Optional[Union[Path, str]] = None) -> None:
        """Add a child node to the current node's children.

        :param node: CoverageNode to add as a child.
        :param path: Relative path to the parent of the child to add.
        """
        self._summary = None

        if dirpath is None:
            if node.name in self._children:
                raise ValueError(f"A node with the name {node.name} was already added as a child")
            self._children[node.name] = node
            return

        path = Path(dirpath)

        if len(path.parts) == 0:
            raise ValueError(f"Invalid child dirpath: '{path}'")

        directory_name = path.parts[0]
        remaining_path = Path(*path.parts[1:]) if len(path.parts) > 1 else None

        if directory_name not in self._children:
            directory_node = CoverageNode(directory_name, node_type=CoverageNodeType.DIR)
            self._children[directory_name] = directory_node

        self._children[directory_name].add_child(node, dirpath=remaining_path)

    @property
    def summary(self) -> CoverageSummary:
        """Get a CoverageSummary for the node.

        :return: CoverageSummary for the node.
        """
        if self._summary is not None:
            return self._summary

        summary = CoverageSummary(0, 0, 0, 0)
        for child in self.children():
            summary.n_lines += child.summary.n_lines
            summary.n_lines_covered += child.summary.n_lines_covered
            summary.n_branches += child.summary.n_branches
            summary.n_branches_covered += child.summary.n_branches_covered
        self._summary = summary
        return self._summary

    def serialize(self) -> Dict[str, Any]:
        """Serialize a tree of CoverageNodes to a tree of Python dictionaries.

        :return: Dictionary representation of the coverage tree.
        """
        summary = {
            'n_lines': self.summary.n_lines,
            'line_rate': self.summary.line_rate,
            'n_lines_covered': self.summary.n_lines_covered,
            'n_branches': self.summary.n_branches,
            'branch_rate': self.summary.branch_rate,
            'n_branches_covered': self.summary.n_branches_covered,
        }
        children = [child.serialize() for child in self.children()]
        node = {
            'name': self.name,
            'summary': summary,
            'node_type': self.node_type.value,
            'children': children,
        }
        return node

    def children(self) -> Generator['CoverageNode', None, None]:
        """Iterate over CoverageNode children."""
        for _, child in self._children.items():
            yield child
